#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import cvxpy as cp
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.metrics import r2_score

# User
from ..utils.exceptions import WrongInputException

# Read about score function here:
# https://kiwidamien.github.io/custom-loss-vs-custom-scoring.html
# https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html

#------------------------------------------------------------------------------
# Constrained OLS
#------------------------------------------------------------------------------
class ConstrainedOLS(BaseEstimator, RegressorMixin):
    # --------------------
    # Constructor function
    # --------------------
    def __init__(self,
                 coefs_lower_bound = None,
                 coefs_lower_bound_constraint = ">=",
                 coefs_upper_bound = None,
                 coefs_upper_bound_constraint = "<=",
                 coefs_sum_bound = None,
                 coefs_sum_bound_constraint = "<=",
                 verbose = False,
                 ):
        self.coefs_lower_bound = coefs_lower_bound
        self.coefs_lower_bound_constraint = coefs_lower_bound_constraint
        self.coefs_upper_bound = coefs_upper_bound
        self.coefs_upper_bound_constraint = coefs_upper_bound_constraint
        self.coefs_sum_bound = coefs_sum_bound
        self.coefs_sum_bound_constraint = coefs_sum_bound_constraint
        self.verbose = verbose
        
        # Initiate other variables
        self.y_name = ""
        
    # --------------------
    # Class variables
    # --------------------
    LOWER_CONSTRAINTS_ALLOWED = [">=", ">"]
    UPPER_CONSTRAINTS_ALLOWED = ["<=", "<"]
    SUM_CONSTRAINTS_ALLOWED = ["==", ">=", ">","<=", "<"]

    # --------------------
    # Public functions
    # --------------------
    # def get_params(self, deep=True):
    #     pass
    # def set_params(self):
    #     pass

    def fit(self, X, y):
        
        # Check that X and y have correct shape
        X, y = check_X_y(X, y)
        
        if isinstance(y, pd.Series):
            self.y_name = y.name
        
        # Break links and transform
        X = np.array(X.copy())
        y = np.array(y.copy())
        
        # Reshape
        y = y.reshape(-1,)
        
        if X.shape[0]!=y.shape[0]:
            raise Exception(f"y is {y.shape}-dim vector, but X is {X.shape}-dim matrix")
        
        # Set up decision variables, i.e. beta coefficients
        beta = cp.Variable(shape=(X.shape[1],), name="beta", integer=False)

        # Initialize constraint    
        constraints = []

        # Lower constraints        
        if self.coefs_lower_bound is not None:
            if self.coefs_lower_bound_constraint == ">=":
                constraints += [beta >= self.coefs_lower_bound]
            elif self.coefs_lower_bound_constraint == ">":
                constraints += [beta > self.coefs_lower_bound]
            else:
                raise WrongInputException(input_name="coefs_lower_bound_constraint",
                                          provided_input=self.coefs_lower_bound_constraint,
                                          allowed_inputs=self.LOWER_CONSTRAINTS_ALLOWED)
                
        # Upper constraints        
        if self.coefs_upper_bound is not None:
            if self.coefs_upper_bound_constraint == "<=":
                constraints += [beta <= self.coefs_upper_bound]
            elif self.coefs_upper_bound_constraint == "<":
                constraints += [beta < self.coefs_upper_bound]
            else:
                raise WrongInputException(input_name="coefs_upper_bound_constraint",
                                          provided_input=self.coefs_upper_bound_constraint,
                                          allowed_inputs=self.UPPER_CONSTRAINTS_ALLOWED)

        # Sum constraints        
        if self.coefs_sum_bound is not None:
            if self.coefs_sum_bound_constraint == "==":
                constraints += [cp.sum(beta) == self.coefs_sum_bound]
            elif self.coefs_sum_bound_constraint == ">=":
                constraints += [cp.sum(beta) >= self.coefs_sum_bound] 
            elif self.coefs_sum_bound_constraint == ">":
                constraints += [cp.sum(beta) > self.coefs_sum_bound]
            elif self.coefs_sum_bound_constraint == "<=":
                constraints += [cp.sum(beta) <= self.coefs_sum_bound]
            elif self.coefs_sum_bound_constraint == "<":
                constraints += [cp.sum(beta) < self.coefs_sum_bound]
            else:
                raise WrongInputException(input_name="coefs_sum_bound_constraint",
                                          provided_input=self.coefs_sum_bound_constraint,
                                          allowed_inputs=self.SUM_CONSTRAINTS_ALLOWED)
            
        # Set up ojective function
        objective = cp.Minimize(cp.sum_squares(y - X @ beta))
        
        # Instantiate and solve
        optim_problem = cp.Problem(objective=objective, constraints=constraints).solve(verbose=self.verbose)
        
        # Beta hat
        self.beta_ = beta.value
        
        return self
        
        
    def predict(self, X):
        
        # Check is fit had been called
        check_is_fitted(self)
        
        # Input validation
        X = check_array(X)
        
        # Break links and transform
        X = np.array(X.copy())
        
        y_hat = pd.Series(X @ self.beta_, name=self.y_name)

        return y_hat

    def score(self, X, y, sample_weight=None):

        y_pred = self.predict(X)
        return r2_score(y, y_pred, sample_weight=sample_weight)
        


#------------------------------------------------------------------------------
# Run interactively
#------------------------------------------------------------------------------
"""
If running program interactively, please uncomment the lines below.
But first, change the 'path_to_here' variable to your local folder!
"""
# import os
# # Manually set path of current file
# path_to_here = "/Users/muhlbach/Repositories/treebased_synthetic_controls/treebased_synthetic_controls/"
# # Change path
# os.chdir(path_to_here)
#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import pandas as pd
import numpy as np
import time

# User
from utils import data
from synthetic_controls import SyntheticControl as SC
from synthetic_controls import TreeBasedSyntheticControl as TBSC
from synthetic_controls import ElasticNetSyntheticControl as ENSC

#------------------------------------------------------------------------------
# Simple example
#------------------------------------------------------------------------------
# Simulation settings
sim_settings = {
    # General    
    "T0":500,
    "T1":50,
    "ate":1,
        
    # Y=f*
    "f":data.generate_linear_data, # generate_linear_data, generate_friedman_data_1, generate_friedman_data_2
    "beta":1,
    "include_intercept":False,
    "expand":False,
    "degree":2,
    "interaction_only":True,

    # Errors
    "eps_mean":0,
    "eps_std":1,
    "eps_cov_x":0,
    "eps_cov_y":0,
    
    # X
    "X_type":"iid",
    "X_dim":5,
    "AR_lags":3,
    }

# Start timer
t0 = time.time()

# Generate data
df = data.simulate_data(**sim_settings)

# Extract data
Y = df["Y"]
W = df["W"]
X = df[[col for col in df.columns if "X" in col]]

# Instantiate SC-objects
sc = SC()
tbsc = TBSC(max_n_models=5, n_folds=1)
ensc = ENSC(max_n_models=5, n_folds=1)

# Fit
sc.fit(Y=Y,W=W,X=X)
print(f"Estimated ATE using SC: {np.around(sc.average_treatment_effet_,2)}")

tbsc.fit(Y=Y,W=W,X=X)
print(f"Estimated ATE using TB-SC: {np.around(tbsc.average_treatment_effet_,2)}")

ensc.fit(Y=Y,W=W,X=X)
print(f"Estimated ATE using EN-SC: {np.around(ensc.average_treatment_effet_,2)}")

# # Bootstrap
# bootstrapped_results = tbsc.bootstrap_ate()

# Stop timer
t1 = time.time()

#------------------------------------------------------------------------------
# The End
#------------------------------------------------------------------------------
print(f"""**************************************************\nCode finished in {np.round(t1-t0, 1)} seconds\n**************************************************""")





#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Main
#------------------------------------------------------------------------------
# class CateError(Exception):
#     """
#     Exception raised when heterogeneous treatment effects are not available
#     """
#     def __init__(self, message="Heterogeneous treatment effects are not available."):
#         self.message = message
#         super().__init__(self.message)    

class WrongInputException(Exception):
    """
    Exception raised when input is not a pd.DataFrame
    """
    def __init__(self, input_name, provided_input, allowed_inputs, message="\nUser-supplied argument '{0}' is currently '{1}'.\nArgument must be one of: '{2}'"):
        self.input_name = input_name
        self.provided_input = provided_input
        self.allowed_inputs = allowed_inputs
        self.message = message
        super().__init__(self.message)    

    def __str__(self):
        
        return self.message.format(self.input_name,self.provided_input, self.allowed_inputs)



#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import numpy as np
import pandas as pd
from sklearn.preprocessing import PolynomialFeatures
from statsmodels.tools.tools import add_constant

###############################################################################
# Main
###############################################################################

#------------------------------------------------------------------------------
# Tools
#------------------------------------------------------------------------------
def get_colnames(x,prefix="X"):
    try:
        dim = x.shape[1]
        colnames = [prefix+str(j) for j in np.arange(start=1,stop=dim+1)]
    except IndexError:
        colnames = [prefix]
        
    return colnames

def convert_to_dict_series(Yobs=None,Ytrue=None,Y0=None,Y1=None,W=None):
    # Save local arguments
    args = locals()
    
    # Convert values to series with appropriate names
    args = {k: pd.Series(v, name=k) for k,v in args.items() if v is not None}
    
    return args

def convert_to_dict_df(X=None):
    # Save local arguments
    args = locals()
    
    # Convert values to series with appropriate names
    args = {k: pd.DataFrame(v, columns=get_colnames(x=v,prefix=k)) for k,v in args.items() if v is not None}
    
    return args

#------------------------------------------------------------------------------
# Generate X data
#------------------------------------------------------------------------------
def generate_ar_process(T=100, x_p=5, ar_p=3, burnin=50, **kwargs):

    # Extract/generate initial coeffients of X
    mu = kwargs.get('mu', 0)
    sigma = kwargs.get('sigma', 1)

    ## Extract/generate parameters for AR
    const = kwargs.get('const', 0)
    ar_coefs = kwargs.get('ar_coefs', np.linspace(start=0.5, stop=0, num=ar_p, endpoint=False))
    error_coef = kwargs.get('error_coef', 1)

    # Fix AR coefs; flip order and reshape to comformable shape
    ar_coefs = np.flip(ar_coefs).reshape(-1,1)

    # Generate errors
    errors = kwargs.get('errors', np.random.multivariate_normal(mean=np.ones(x_p), 
                                                                cov=np.identity(x_p),
                                                                size=T))    

    # Generate errors for burn-in period
    errors_burnin = np.random.multivariate_normal(mean=np.mean(errors,axis=0), 
                                                  cov=np.cov(errors.T),
                                                  size=burnin)

    errors_all = np.concatenate((errors_burnin,errors))

    # Generate initial value(s)
    X = mu + sigma * np.random.randn(ar_p,x_p)

    # Simulate AR(p) with burn-in included
    for b in range(burnin+T):
        X = np.concatenate((X,
                            const + ar_coefs.T @ X[0:ar_p,:] + error_coef * errors_all[b,0:x_p]),
                           axis=0)

    # Return only the last T observations (we have removed the dependency on the initial draws)
    return X[-T:,]

def generate_iid_process(T=100, x_p=5, **kwargs):

    # Extract/generate initial coeffients of X
    mu = kwargs.get('mu', 0)
    sigma = kwargs.get('sigma', 1)
    covariance = kwargs.get('covariance', 0)

    # Construct variance-covariance matrix
    cov_diag = np.diag(np.repeat(a=sigma**2, repeats=x_p))
    cov_off_diag= np.ones(shape=(x_p,x_p)) * covariance
    np.fill_diagonal(a=cov_off_diag, val=0)
    cov_mat = cov_diag + cov_off_diag

    # Generate X
    X = np.random.multivariate_normal(mean=np.repeat(a=mu, repeats=x_p), 
                                      cov=cov_mat,
                                      size=T)    

    return X

def generate_errors(N=1000, p=5, mu=0, sigma=1, cov_X=0.25, cov_y=0.5):

    # Number of dimensions including y
    n_dim = p+1

    ## Construct variance-covariance matrix
    # Construct diagonal with variance = sigma^2
    cov_diag = np.diag(np.repeat(a=sigma**2, repeats=n_dim))
    
    ## Construct off-diagonal with covariances
    # Fill out for X (and y)
    cov_off_diag = np.ones(shape=(n_dim,n_dim)) * cov_X
    
    # Update y entries
    cov_off_diag[p,:] = cov_off_diag[:,p] = cov_y
    
    # Set diagonal to zero
    np.fill_diagonal(a=cov_off_diag, val=0)
    
    # Update final variance-covariance matrix
    cov_mat = cov_diag + cov_off_diag

    # Generate epsilon
    eps = np.random.multivariate_normal(mean=np.repeat(a=mu, repeats=n_dim), 
                                        cov=cov_mat,
                                        size=N)    

    return eps

#------------------------------------------------------------------------------
# Generate f_star = E[Y|X=x]
#------------------------------------------------------------------------------
def generate_linear_data(x, beta=1, include_intercept=True, expand=True, degree=2, interaction_only=True, **kwargs):

    # Convert to np and break link
    x = np.array(x.copy())    

    if expand:
        
        # Instantiate 
        polynomialfeatures = PolynomialFeatures(degree=degree, interaction_only=interaction_only, include_bias=False, order='C')
    
        # Expand x
        x = polynomialfeatures.fit_transform(x)
    
    if include_intercept:
        x = add_constant(data=x, prepend=True, has_constant='skip')
    
    if isinstance(beta, int) or isinstance(beta, float):
        beta = np.repeat(a=beta, repeats=x.shape[1])
    
    # Make sure beta has the right dimensions
    beta = beta.reshape(-1,1)
    
    if x.shape[1]!=beta.shape[0]:
        raise Exception(f"Beta is {beta.shape}-dim vector, but X is {x.shape}-dim matrix")

    # Generate fstar=E[y|X=x]
    f_star = x @ beta
    
    # Reshape for conformity
    f_star = f_star.reshape(-1,)
    
    return f_star
    

def generate_friedman_data_1(x, **kwargs):
    
    # Convert to np and break link
    x = np.array(x.copy())    

    # Generate fstar=E[y|X=x]
    f_star = 0.1*np.exp(4*x[:,0]) + 4/(1+np.exp(-20*(x[:,1]-0.5))) + 3*x[:,2] + 2*x[:,3] + 1*x[:,4]
    
    # Reshape for conformity
    f_star = f_star.reshape(-1,)
    
    return f_star

def generate_friedman_data_2(x, **kwargs):
    
    # Convert to np and break link
    x = np.array(x.copy())    

    # Generate fstar=E[y|X=x]
    f_star = 10*np.sin(np.pi*x[:,0]*x[:,1]) + 20*(x[:,2]-0.5)**2 + 10*x[:,3] + 5*x[:,4]
    
    # Reshape for conformity
    f_star = f_star.reshape(-1,)
    
    return f_star

#------------------------------------------------------------------------------
# Simulate data
#------------------------------------------------------------------------------
def simulate_data(f,
                  T0=500,
                  T1=50,
                  X_type="AR",
                  X_dim=5,
                  AR_lags=3,
                  ate=1,
                  eps_mean=0,
                  eps_std=1,
                  eps_cov_x=0,
                  eps_cov_y=0,
                  **kwargs):

    # Total number of time periods
    T = T0 + T1

    # Generate errors
    errors = generate_errors(N=T, p=X_dim, mu=eps_mean, sigma=eps_std, cov_X=eps_cov_x, cov_y=eps_cov_y)
    
    # Generate covariates
    if X_type=="AR":
        X = generate_ar_process(T=T,
                                x_p=X_dim,
                                ar_p=AR_lags,
                                errors=errors)
                
    elif X_type=="iid":
        X = generate_iid_process(T=T,x_p=X_dim)
    
    # Generate W
    W = np.repeat((0,1), (T0,T1))

    # Generate Y
    Y = f(x=X, **kwargs) + ate*W + errors[:,-1]

    # df = {"Y":Y,
    #       "W":W,
    #       "X":X}

    # Collect data
    df = pd.concat(objs=[pd.Series(data=Y,name="Y"),
                          pd.Series(data=W,name="W"),
                          pd.DataFrame(data=X,columns=[f"X{d}" for d in range(X.shape[1])])],
                    axis=1)
        
    return df

def generate_data(dgp="AR1", ar_p=1, n_controls=5, T0=500, T1=50, return_as_df=False, **kwargs):
    
    # Valid dgps
    VALID_DGP = ["AR1"]
    
    # Check if dgp is valid
    if not dgp in VALID_DGP:
        raise Exception(f"Choice of input 'dgp' must be one of {VALID_DGP}, but is currently 'dgp'")
    
    # Total number of time periods
    T = T0 + T1
    
    # Generate data
    if dgp=="AR1":
        
        # Number of control units (columns of X)
        x_p=5
        
        # Numbers of lags in AR
        ar_p=2
        
        # Coefficients in AR
        ar_coefs=np.array([0.5, 0.25])
        
        # Errors of X
        errors = np.random.multivariate_normal(mean=np.ones(x_p), 
                                               cov=np.diag(np.ones(x_p)),
                                               size=T)
        
        # Generate X
        X = generate_ar_process(T=T,
                                x_p=x_p,
                                ar_p=ar_p,
                                ar_coefs=ar_coefs,
                                errors=errors)
        
        # beta coefficients as in X*beta
        beta = kwargs.get('beta', np.ones(x_p))


        # tau as in treatment effect Y1-Y0
        tau = kwargs.get('tau', 5)

        # Covariance of eps
        Gamma = kwargs.get('Gamma', np.identity(2))
        
        # Error term
        epsilon = np.random.multivariate_normal(mean=np.zeros(2), cov=Gamma, size=T)

        # Treatment dummy
        W = np.concatenate((np.repeat(0, repeats=T0), np.repeat(1, repeats=T1)), axis=0)
        
        # Potential outcomes
        Y_baseline = X @ beta
        Y0 = Y_baseline + epsilon[:,0]
        Y1 = tau + Y_baseline + epsilon[:,1]
        
        # Observed outcome
        Y_obs = (1-W)*Y0 + W*Y1
    
        # Transform data
        data_output = convert_to_dict_series(Yobs=Y_obs,Ytrue=Y_baseline,Y0=Y0,Y1=Y1,W=W)
        data_input = convert_to_dict_df(X=X)
        
        # Return as df
        df = pd.concat([pd.DataFrame().from_records(data_output),
                        data_input["X"]], axis=1)
        
        # House-keeping
        # del X,Y_obs,Y_baseline,Y0,Y1,W,data_output,data_input
        
        return df
        
    
                                  


#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import numpy as np
import pandas as pd
import math

# User
from .exceptions import WrongInputException
from .tools import break_links
#------------------------------------------------------------------------------
# Non-balanced IID bootstrap
#------------------------------------------------------------------------------
class Bootstrap(object):
    """
    Various bootstraps
    """
    # --------------------
    # Constructor function
    # --------------------
    def __init__(self,
                 bootstrap_type="iid",
                 n_bootstrap_samples=100,
                 block_length=5,
                 ):
        # Initialize inputs
        self.bootstrap_type = bootstrap_type
        self.n_bootstrap_samples = n_bootstrap_samples
        self.block_length = block_length

    # --------------------
    # Class variables
    # --------------------
    BOOTSTRAP_TYPE_ALLOWED = ["iid", "circular"]
    X_ALLOWED = ["pd.Series", "pd.DataFrame", "np.ndarray"]
    
    # --------------------
    # Private functions
    # --------------------

    # --------------------
    # Public functions
    # --------------------
    def generate_samples(self, x):

        x = break_links(x)        
        
        if isinstance(x, pd.Series):
            pass
        elif isinstance(x, pd.DataFrame):
            x = x.squeeze()
        elif isinstance(x, np.ndarray):
            x = pd.Series(x)
        else:
            raise WrongInputException(input_name="x", provided_input=x, allowed_inputs=self.X_ALLOWED)
        
        # Get size
        n_obs = x.shape[0]
        
        # Save original indices
        original_indices = x.index
        
        # Reset index
        x.reset_index(drop=True,inplace=True)
        
        if self.bootstrap_type=="iid":        
            # Generate random integer indices from 0 to n-1
            indices = np.random.randint(low=0, high=n_obs, size=(n_obs, self.n_bootstrap_samples))
            
        elif self.bootstrap_type=="circular":
            # Stack observations such that all blocks have the same option of being chosen
            x = pd.concat([x, x.iloc[0:(self.block_length-1)]], axis=0).reset_index(drop=True)
            
            # 
            block_start_indices = np.array(range(n_obs))            
            add_indices  = np.arange(self.block_length, dtype=int).reshape((1, self.block_length))
            
            indices = np.random.choice(block_start_indices,
                                       size=(self.n_bootstrap_samples,math.ceil(n_obs / self.block_length),1),
                                       replace=True)
            
            # Add successive indices
            indices = (indices + add_indices)

            # Reshape
            indices = indices.ravel().reshape((n_obs,-1), order="F")
        
        # Sample
        x_resampled = pd.DataFrame(np.array(x)[indices])

        # Re-index
        x_resampled.index = original_indices
        
        return x_resampled

    
        


#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
import pandas as pd
import numpy as np
from collections.abc import Sequence
#------------------------------------------------------------------------------
# Main
#------------------------------------------------------------------------------
def check_estimator(estimator):
    """Check whether model instance is useful for estimating propensity scores"""
    # Methods required
    required_methods = ["fit", "predict","get_params", "set_params"]

    # Check if class has required methods by using getattr() to get the attribute, and callable() to verify it is a method
    for method in required_methods:
        method_attribute = getattr(estimator, method, None)
        if not callable(method_attribute):
            raise Exception(f"Estimator '{type(estimator).__name__}' does not contain the method '{method}'")

def check_param_grid(param_grid):
    if hasattr(param_grid, "items"):
        param_grid = [param_grid]

    for p in param_grid:
        for name, v in p.items():
            if isinstance(v, np.ndarray) and v.ndim > 1:
                raise ValueError("Parameter array should be one-dimensional.")

            if isinstance(v, str) or not isinstance(v, (np.ndarray, Sequence)):
                raise ValueError(
                    "Parameter grid for parameter ({0}) needs to"
                    " be a list or numpy array, but got ({1})."
                    " Single values need to be wrapped in a list"
                    " with one element.".format(name, type(v))
                )

            if len(v) == 0:
                raise ValueError(
                    "Parameter values for parameter ({0}) need "
                    "to be a non-empty sequence.".format(name)
                )
                
def check_X(X):
    
    # Convert to dataframe if not already converted
    if not isinstance(X, pd.DataFrame):
        X = pd.DataFrame(X)
    
    # Check for numeric dtype
    if not all(X.dtypes.map(pd.api.types.is_numeric_dtype)):
        raise Exception("X must be only numeric dtype")
    
    # Check for missing
    if X.isnull().values.any():
        raise Exception("X contains missing inputs")
        
    return X
                
def check_Y(Y):
    
    # Convert to series if not already converted
    if not isinstance(Y, pd.Series):
        Y = pd.Series(Y)
    
    # Check for numeric dtype
    if not pd.api.types.is_numeric_dtype(Y.dtype):
        raise Exception("Y must be numeric dtype")
    
    # Check for missing
    if Y.isnull().values.any():
        raise Exception("Y contains missing inputs")        
        
    return Y
        
def check_X_Y(X, Y):
    
    # Perform individual checks
    X = check_X(X)
    Y = check_Y(Y)
    
    # Check length
    if len(Y)!=len(X):
        raise Exception(f"Y has {len(Y)} observations, whereas X has {len(X)} observations. They must match.")

    return X, Y


#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import ElasticNet
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV, KFold, TimeSeriesSplit

# User
from .base.base_synthetic_controls import BaseSyntheticControl
from .estimator.constrained_ols import ConstrainedOLS
from .utils.sanity_check import check_param_grid
from .utils.tools import SingleSplit
from .utils.exceptions import WrongInputException
from .utils.bootstrap import Bootstrap

#------------------------------------------------------------------------------
# Ordinary Synthetic Control Group Method
#------------------------------------------------------------------------------
class SyntheticControl(BaseSyntheticControl):
    """
    This class estimates the average treatment effects by constructing a synthetic control group using constrained OLS
    """
    # --------------------
    # Constructor function
    # --------------------
    # HERE!
    def __init__(self,
                 estimator=ConstrainedOLS(),
                 param_grid={'coefs_lower_bound':0,
                             'coefs_lower_bound_constraint':">=",
                             'coefs_sum_bound':1,
                             'coefs_sum_bound_constraint':"<=",},
                 cv_params={'scoring':None,
                            'n_jobs':None,
                            'refit':True,
                            'verbose':0,
                            'pre_dispatch':'2*n_jobs',
                            'random_state':None,
                            'error_score':np.nan,
                            'return_train_score':False},
                 n_folds=3,
                 fold_type="KFold",
                 max_n_models=50,
                 test_size=0.25,
                 verbose=False,
                 ):
        super().__init__(
            estimator=estimator,
            param_grid=param_grid,
            cv_params=cv_params,
            n_folds=n_folds,
            fold_type=fold_type,
            max_n_models=max_n_models,
            test_size=test_size,
            verbose=verbose,
            )



#------------------------------------------------------------------------------
# Tree-Based Synthetic Control Group Method
#------------------------------------------------------------------------------
class TreeBasedSyntheticControl(BaseSyntheticControl):
    """
    This class estimates the average treatment effects by constructing a synthetic control group using Random Forests
    """
    # --------------------
    # Constructor function
    # --------------------
    def __init__(self,
                 estimator=RandomForestRegressor(criterion='mse',
                                                 min_weight_fraction_leaf=0.0,
                                                 min_impurity_decrease=0.0,
                                                 bootstrap=False,
                                                 oob_score=False,
                                                 n_jobs=None,
                                                 random_state=None,
                                                 verbose=0,
                                                 warm_start=False,
                                                 ccp_alpha=0.0,
                                                 max_samples=None),
                 param_grid={'n_estimators': 500,
                             'max_depth': [2,4,8,16,None],
                             'min_samples_split': [2,4,8,16],
                             'min_samples_leaf': [1,2,4,8],
                             'max_features': [1/4,1/3,1/2,2/3, 'sqrt','log2'],
                             'max_leaf_nodes': None,
                             },
                 cv_params={'scoring':None,
                            'n_jobs':None,
                            'refit':True,
                            'verbose':0,
                            'pre_dispatch':'2*n_jobs',
                            'random_state':None,
                            'error_score':np.nan,
                            'return_train_score':False},
                 n_folds=3,
                 fold_type="KFold",
                 max_n_models=50,
                 test_size=0.25,
                 verbose=False,
                 ):
        super().__init__(
            estimator=estimator,
            param_grid=param_grid,
            cv_params=cv_params,
            n_folds=n_folds,
            fold_type=fold_type,
            max_n_models=max_n_models,
            test_size=test_size,
            verbose=verbose,
            )

            
#------------------------------------------------------------------------------
# Elastic Net Synthetic Control Group Method
#------------------------------------------------------------------------------
class ElasticNetSyntheticControl(BaseSyntheticControl):
    """
    This class estimates the average treatment effects by constructing a synthetic control group using Elastic Net
    """
    # --------------------
    # Constructor function
    # --------------------
    def __init__(self,
                 estimator=ElasticNet(fit_intercept=True,
                                      precompute=False,
                                      max_iter=10000,
                                      copy_X=True,
                                      tol=0.0001,
                                      warm_start=False,
                                      positive=False,
                                      random_state=None,
                                      selection='cyclic'),
                 param_grid={"l1_ratio": [1/10, 1/4, 1/2, 3/4, 1],
                             "alpha": np.exp(np.linspace(start=np.log(100), stop=np.log(0.000001), num=100)),
                             },
                 cv_params={'scoring':None,
                            'n_jobs':None,
                            'refit':True,
                            'verbose':0,
                            'pre_dispatch':'2*n_jobs',
                            'random_state':None,
                            'error_score':np.nan,
                            'return_train_score':False},
                 n_folds=3,
                 fold_type="KFold",
                 max_n_models=50,
                 test_size=0.25,
                 verbose=False,
                 ):
        super().__init__(
            estimator=estimator,
            param_grid=param_grid,
            cv_params=cv_params,
            n_folds=n_folds,
            fold_type=fold_type,
            max_n_models=max_n_models,
            test_size=test_size,
            verbose=verbose,
            )
    
    
        


#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import numpy as np
import pickle
from sklearn.utils import indexable
from sklearn.utils.validation import _num_samples

#------------------------------------------------------------------------------
# Main
#------------------------------------------------------------------------------
# Save and load objects using pickle
def save_object_by_pickle(path,obj):
    with open(path, 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

def load_object_by_pickle(path):
    with open(path, 'rb') as f:
        return pickle.load(f)

def break_links(x):
    """ Break link to x by copying it"""
    x = x.copy()
    return x

def update_parameters(self, default_params, user_params, element_as_list=True):
    """ Merge default and supplied parameters """
    if bool(user_params):
        for k,v in user_params.items():
            
            # If key is in param_default, then update value
            if k in default_params:
                default_params[k] = v
            else:
                raise Exception(f"Parameter '{k}' is invalid because it is not part of default parameters")
            
    if element_as_list:
        # Turn every element into a list    
        default_params = {key:(value if isinstance(value, list) else [value]) for key,value in default_params.items()}
            
    return default_params    


class SingleSplit(object):
    # --------------------
    # Constructor function
    # --------------------
    def __init__(self,
                 test_size=0.25):
        self.test_size=test_size

    # --------------------
    # CLASS VARIABLES
    # --------------------        
    N_SPLITS = 1
        
    # --------------------
    # Public
    # --------------------        
    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        y : array-like of shape (n_samples,), default=None
            The target variable for supervised learning problems.
        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into
            train/test set.
        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """
        X, y, groups = indexable(X, y, groups)
        n_samples = _num_samples(X)    
        indices = np.arange(n_samples)

        n_test = int(self.test_size*n_samples)
        n_train = n_samples-n_test
        
        train = indices[0:n_train]
        test  = indices[n_train:n_samples]
        
        yield train, test

    def get_n_splits(self, X=None, y=None, groups=None):
        return self.N_SPLITS
